unit chargraf;

{ chargraf.pas }
{ unit for chart.pas }

{ (c) Heiko Purnhagen 05-aug-95 }
{ using setwritemode for revline   HP 960725 }
{ added revcircle   HP 960726 }

interface

uses
  graph,crt,
  mousedrv,grafbmp;

const
  mouseratio = 4;   { mouse fine speed-ratio }

  curkey = 20;

  curup = #72;     { up }
  curdown = #80;   { down }
  curleft = #75;   { left }
  curright = #77;  { right }
  keyup = #5;      { ^E }
  keydown = #24;   { ^X }
  keyleft = #19;   { ^S }
  keyright = #4;   { ^D }

  maxmenu = 10;

  tx1 = 2;
  tx2 = 1;
  ty1 = 2;
  ty2 = 1;
  txb = 2;
  twb = 10;
  lowcur = 1;

  wincol = 70;
  winrow = 9;

  esc = #27;
  ctrly = #25;
  backsp = #8;
  cr = #13;
  tab = #9;
  cup = #72;
  cdown = #80;
  cleft = #75;
  cright = #77;
  cpgup = #73;
  cpgdn = #81;
  chome = #71;
  cend = #79;
  cdel = #83;
  cins = #82;

  circleseg4 = 4;


type
  string20 = string[20];
  menut = record
            x1,y1,x2,y2 : integer;
            vertical : boolean;
            enum : integer;
            eheight : integer;
            ewidth : array [1..maxmenu] of integer;
            key : array [1..maxmenu] of char;
            name : array [1..maxmenu] of string20;
            background : pointer;
          end;


var
  xmax,ymax : integer;
  xasp,yasp : word;
  reverse : boolean;
  bgc,fgc : integer;
  mousefine : boolean;
  mouseslow : boolean;
  cursorbig : boolean;
  usemouse : boolean;

  mx,my,mxa,mya : integer;
  mbl,mbr,mbla,mbra : boolean;
  mblold,mbrold : boolean;
  cursorar : array [0..3] of cursort;
  cxi,cyi : integer;
  cx,cy : real;
  mmouseratio : integer;
  mmousefine : boolean;
  mkey : boolean;
  mc1,mc2 : char;

  menum : menut;
  menumaxwidth,menumaxsize : integer;

  th,tw : integer;
  winx1,winy1,winx2,winy2 : integer;
  winxc : integer;
  winyr : array [1..winrow] of integer;
  winbackground : pointer;
  confirmbackground : pointer;

  redraw,recalc : boolean;

  circlesin : array[0..circleseg4*4-1] of real;


procedure revpixel (x,y : integer);
procedure revline (x1,y1,x2,y2 : integer);
procedure revcircleinit;
procedure revcircle (x,y,rx,ry : integer);
procedure drawmarkdot (x,y : integer);
procedure drawmarkcircle (x,y : integer);
procedure newscreen;
procedure box (x1,y1,x2,y2 : integer);
procedure revbox (x1,y1,x2,y2 : integer);
procedure errorbeep;

procedure initfinecolor;
procedure initfine;
procedure xhidecursor;
procedure xshowcursor;
procedure startfine;
procedure updatefine (usecurkeys : boolean);
procedure endfine;
function clickedl : boolean;
function clickedr : boolean;
function anyclick : boolean;
function getmkey : char;
function clickedlkey : boolean;
function clickedrkey : boolean;
procedure waitclickkey;

procedure newtext (x,y : integer; xcent : boolean; s : string);
procedure boxtext (x1,y1,x2,y2 : integer; xcent : boolean; s : string);
function editchar (x,y : integer; que : string; var ans : char) : boolean;
function editline (x,y : integer; que : string; var ans : string) : boolean;
procedure initm (x,y : integer; vert : boolean);
procedure nextm (c : char; s : string20);
procedure quitm (var mm : menut);
procedure drawmenu (var m : menut);
function checkmenu (var m : menut) : char;
procedure freemenu (var m : menut);
procedure initconfirm (minleft : longint);
function confirm (s : string; yesno : boolean) : boolean;
procedure initwindow (minleft : longint);
procedure openwindow;
procedure closewindow;


implementation

{ reverse line with bresenham }

procedure revpixel (x,y : integer);
begin
  putpixel(x,y,getmaxcolor-getpixel(x,y));
end;

procedure revline1 (x,y,dbig,dsmall,dir : integer);
  var
    i,j,d : integer;
begin
  if dir > 4 then
    d := (dbig+1) div 2
  else
    d := dbig div 2;
  j := 0;
  for i := 0 to dbig do
  begin
    case dir of
      1 : revpixel (x+i,y+j);
      2 : revpixel (x+j,y+i);
      3 : revpixel (x-j,y+i);
      4 : revpixel (x-i,y+j);
      5 : revpixel (x-i,y-j);
      6 : revpixel (x-j,y-i);
      7 : revpixel (x+j,y-i);
      8 : revpixel (x+i,y-j);
    end;
    d := d + dsmall;
    if d >= dbig then
    begin
      inc(j);
      d := d - dbig;
    end;
  end;
end;

procedure revline_old (x1,y1,x2,y2 : integer);
{ somehow buggy ... (vert. line len=2)   HP 960725 }
  var
    dx,dy : integer;
begin
  dx := x2-x1;
  dy := y2-y1;
  if dx >= 0 then
  begin
    if dy >= 0 then
    begin
      if dx >= dy then
        revline1 (x1,y1,dx,dy,1)
      else
        revline1 (x1,y1,dy,dx,2);
    end
    else
    begin
      if dx >= -dy then
        revline1 (x1,y1,dx,-dy,8)
      else
        revline1 (x1,y1,-dy,dx,7);
    end;
  end
  else
  begin
    if dy >= 0 then
    begin
      if -dx >= dy then
        revline1 (x1,y1,-dx,dy,4)
      else
        revline1 (x1,y1,dy,-dx,3);
    end
    else
    begin
      if -dx >= -dy then
        revline1 (x1,y1,-dx,-dy,5)
      else
        revline1 (x1,y1,-dy,-dx,6);
    end;
  end;
end;

procedure revline (x1,y1,x2,y2 : integer);
  var
    oldcolor : integer;
begin
  setwritemode (xorput);
  oldcolor := getcolor;
  setcolor (getmaxcolor);
  line (x1,y1,x2,y2);
  setcolor (oldcolor);
  setwritemode (normalput);
end;

procedure revcircleinit;
  var i : integer;
begin
  for i := 0 to circleseg4*4-1 do
    circlesin[i] := sin(i/(circleseg4*4)*2*pi);
end;

procedure revcircle (x,y,rx,ry : integer);
  var
    oldcolor : integer;
    i : integer;
begin
  setwritemode (xorput);
  oldcolor := getcolor;
  setcolor (getmaxcolor);
  moveto (x+round(rx*circlesin[circleseg4*4-1]),
          y+round(ry*circlesin[circleseg4-1]));
  for i := 0 to circleseg4*4-1 do
    lineto (x+round(rx*circlesin[i]),
            y+round(ry*circlesin[(circleseg4+i) mod (circleseg4*4)]));
  setcolor (oldcolor);
  setwritemode (normalput);
end;

procedure drawmarkdot (x,y : integer);
begin
  revline (x-1,y-2,x+1,y-2);
  revline (x-2,y-1,x+2,y-1);
  revline (x-2,y,x+2,y);
  revline (x-2,y+1,x+2,y+1);
  revline (x-1,y+2,x+1,y+2);
end;

procedure drawmarkcircle (x,y : integer);
  procedure circlequad (a,b1,b2 : integer);
  begin
    revline (x+b1,y+a,x+b2,y+a);
    revline (x+a,y-b1,x+a,y-b2);
    revline (x-b1,y-a,x-b2,y-a);
    revline (x-a,y+b1,x-a,y+b2);
  end;
begin
  circlequad (6,-4,-3);
  circlequad (8,-2,2);
  circlequad (7,-4,4);
  circlequad (6,3,6);
  circlequad (5,5,6);
end;


{ procedures for mouse cursor }

procedure setcursorar;

  procedure setcurline (var cur : cursort; line : integer; s : string);

  { cursor[line] cursor[lin+16]  code  color       }
  { 0            0               0     black       }
  { 1            0               _     transparent }
  { 0            1               1     white       }
  { 1            1               x     invert      }

    var
      i : integer;
      w : word;
  begin
    cur[line] := 0;
    cur[line+16] := 0;
    w := 1;
    for i := 16 downto 1 do
    begin
      case s[i] of
        '0' : if reverse then
                cur[line+16] := cur[line+16]+w;
        '_' : cur[line] := cur[line]+w;
        '1' : if not reverse then
                cur[line+16] := cur[line+16]+w;
        'x' : begin
                cur[line] := cur[line]+w;
                cur[line+16] := cur[line+16]+w;
              end;
      end;
      w := w*2;
    end;
  end;

begin
  if cursorbig then
  begin
    setcurline (cursorar[0], 0,'_____00000______');
    setcurline (cursorar[0], 1,'_____01110______');
    setcurline (cursorar[0], 2,'_____01110______');
    setcurline (cursorar[0], 3,'_____01110______');
    setcurline (cursorar[0], 4,'_____01110______');
    setcurline (cursorar[0], 5,'000000000000000_');
    setcurline (cursorar[0], 6,'011110___011110_');
    setcurline (cursorar[0], 7,'011110___011110_');
    setcurline (cursorar[0], 8,'011110___011110_');
    setcurline (cursorar[0], 9,'000000000000000_');
    setcurline (cursorar[0],10,'_____01110______');
    setcurline (cursorar[0],11,'_____01110______');
    setcurline (cursorar[0],12,'_____01110______');
    setcurline (cursorar[0],13,'_____01110______');
    setcurline (cursorar[0],14,'_____00000______');
    setcurline (cursorar[0],15,'________________');
    setcurline (cursorar[1], 0,'_____000000_____');
    setcurline (cursorar[1], 1,'_____011110_____');
    setcurline (cursorar[1], 2,'_____011110_____');
    setcurline (cursorar[1], 3,'_____011110_____');
    setcurline (cursorar[1], 4,'_____011110_____');
    setcurline (cursorar[1], 5,'0000000000000000');
    setcurline (cursorar[1], 6,'011110____011110');
    setcurline (cursorar[1], 7,'011110____011110');
    setcurline (cursorar[1], 8,'011110____011110');
    setcurline (cursorar[1], 9,'0000000000000000');
    setcurline (cursorar[1],10,'_____011110_____');
    setcurline (cursorar[1],11,'_____011110_____');
    setcurline (cursorar[1],12,'_____011110_____');
    setcurline (cursorar[1],13,'_____011110_____');
    setcurline (cursorar[1],14,'_____000000_____');
    setcurline (cursorar[1],15,'________________');
    setcurline (cursorar[2], 0,'_____00000______');
    setcurline (cursorar[2], 1,'_____01110______');
    setcurline (cursorar[2], 2,'_____01110______');
    setcurline (cursorar[2], 3,'_____01110______');
    setcurline (cursorar[2], 4,'_____01110______');
    setcurline (cursorar[2], 5,'000000000000000_');
    setcurline (cursorar[2], 6,'011110___011110_');
    setcurline (cursorar[2], 7,'011110___011110_');
    setcurline (cursorar[2], 8,'011110___011110_');
    setcurline (cursorar[2], 9,'011110___011110_');
    setcurline (cursorar[2],10,'000000000000000_');
    setcurline (cursorar[2],11,'_____01110______');
    setcurline (cursorar[2],12,'_____01110______');
    setcurline (cursorar[2],13,'_____01110______');
    setcurline (cursorar[2],14,'_____01110______');
    setcurline (cursorar[2],15,'_____00000______');
    setcurline (cursorar[3], 0,'_____000000_____');
    setcurline (cursorar[3], 1,'_____011110_____');
    setcurline (cursorar[3], 2,'_____011110_____');
    setcurline (cursorar[3], 3,'_____011110_____');
    setcurline (cursorar[3], 4,'_____011110_____');
    setcurline (cursorar[3], 5,'0000000000000000');
    setcurline (cursorar[3], 6,'011110____011110');
    setcurline (cursorar[3], 7,'011110____011110');
    setcurline (cursorar[3], 8,'011110____011110');
    setcurline (cursorar[3], 9,'011110____011110');
    setcurline (cursorar[3],10,'0000000000000000');
    setcurline (cursorar[3],11,'_____011110_____');
    setcurline (cursorar[3],12,'_____011110_____');
    setcurline (cursorar[3],13,'_____011110_____');
    setcurline (cursorar[3],14,'_____011110_____');
    setcurline (cursorar[3],15,'_____000000_____');
  end
  else
  begin
    setcurline (cursorar[0], 0,'______000_______');
    setcurline (cursorar[0], 1,'______010_______');
    setcurline (cursorar[0], 2,'______010_______');
    setcurline (cursorar[0], 3,'______010_______');
    setcurline (cursorar[0], 4,'______010_______');
    setcurline (cursorar[0], 5,'______000_______');
    setcurline (cursorar[0], 6,'000000___000000_');
    setcurline (cursorar[0], 7,'011110___011110_');
    setcurline (cursorar[0], 8,'000000___000000_');
    setcurline (cursorar[0], 9,'______000_______');
    setcurline (cursorar[0],10,'______010_______');
    setcurline (cursorar[0],11,'______010_______');
    setcurline (cursorar[0],12,'______010_______');
    setcurline (cursorar[0],13,'______010_______');
    setcurline (cursorar[0],14,'______000_______');
    setcurline (cursorar[0],15,'________________');
    setcurline (cursorar[1], 0,'______0000______');
    setcurline (cursorar[1], 1,'______0110______');
    setcurline (cursorar[1], 2,'______0110______');
    setcurline (cursorar[1], 3,'______0110______');
    setcurline (cursorar[1], 4,'______0110______');
    setcurline (cursorar[1], 5,'______0000______');
    setcurline (cursorar[1], 6,'000000____000000');
    setcurline (cursorar[1], 7,'011110____011110');
    setcurline (cursorar[1], 8,'000000____000000');
    setcurline (cursorar[1], 9,'______0000______');
    setcurline (cursorar[1],10,'______0110______');
    setcurline (cursorar[1],11,'______0110______');
    setcurline (cursorar[1],12,'______0110______');
    setcurline (cursorar[1],13,'______0110______');
    setcurline (cursorar[1],14,'______0000______');
    setcurline (cursorar[1],15,'________________');
    setcurline (cursorar[2], 0,'______000_______');
    setcurline (cursorar[2], 1,'______010_______');
    setcurline (cursorar[2], 2,'______010_______');
    setcurline (cursorar[2], 3,'______010_______');
    setcurline (cursorar[2], 4,'______010_______');
    setcurline (cursorar[2], 5,'______000_______');
    setcurline (cursorar[2], 6,'000000___000000_');
    setcurline (cursorar[2], 7,'011110___011110_');
    setcurline (cursorar[2], 8,'011110___011110_');
    setcurline (cursorar[2], 9,'000000___000000_');
    setcurline (cursorar[2],10,'______000_______');
    setcurline (cursorar[2],11,'______010_______');
    setcurline (cursorar[2],12,'______010_______');
    setcurline (cursorar[2],13,'______010_______');
    setcurline (cursorar[2],14,'______010_______');
    setcurline (cursorar[2],15,'______000_______');
    setcurline (cursorar[3], 0,'______0000______');
    setcurline (cursorar[3], 1,'______0110______');
    setcurline (cursorar[3], 2,'______0110______');
    setcurline (cursorar[3], 3,'______0110______');
    setcurline (cursorar[3], 4,'______0110______');
    setcurline (cursorar[3], 5,'______0000______');
    setcurline (cursorar[3], 6,'000000____000000');
    setcurline (cursorar[3], 7,'011110____011110');
    setcurline (cursorar[3], 8,'011110____011110');
    setcurline (cursorar[3], 9,'000000____000000');
    setcurline (cursorar[3],10,'______0000______');
    setcurline (cursorar[3],11,'______0110______');
    setcurline (cursorar[3],12,'______0110______');
    setcurline (cursorar[3],13,'______0110______');
    setcurline (cursorar[3],14,'______0110______');
    setcurline (cursorar[3],15,'______0000______');
  end;
end;

procedure revcursor (mx,my : integer);
begin
  revline (mx-7,my-1,mx-4,my-1);
  revline (mx-7,my,mx-4,my);
  revline (mx-7,my+1,mx-4,my+1);
  revline (mx+4,my-1,mx+7,my-1);
  revline (mx+4,my,mx+7,my);
  revline (mx+4,my+1,mx+7,my+1);
  revline (mx-1,my-7,mx-1,my-4);
  revline (mx,my-7,mx,my-4);
  revline (mx+1,my-7,mx+1,my-4);
  revline (mx-1,my+4,mx-1,my+7);
  revline (mx,my+4,mx,my+7);
  revline (mx+1,my+4,mx+1,my+7);
end;

procedure limitborder (var mx,my : integer);
begin
  if mx < 0 then
    mx := 0;
  if mx > xmax then
    mx := xmax;
  if my < 0 then
    my := 0;
  if my > ymax then
    my := ymax;
end;

procedure initfinecolor;
begin
  setcursorar;
  cursor := cursorar[0];
  defgcursor (7,7);
end;

procedure initfine;
begin
  cx := (xmax+1)/2;
  cy := ty1+ty2+textheight(' ')+5+10;
  mx := round(cx);
  my := round(cy);
  if usemouse then
  begin
    setminmaxx (0,xmax);
    setminmaxy (0,ymax);
    setpos (mx,my);
    initfinecolor;
    showcursor;
    getposbut (mx,my,mbl,mbr);
    mbla := mbl;
    mbra := mbr;
    mblold := mbla;
    mbrold := mbra;
    mmouseratio := 1;
    cxi := 0;
    cyi := 0;
  end
  else
    revcursor (mx,my);
  mxa := mx;
  mya := my;
  cxi := mx;
  cyi := my;
  mkey := false;
  mmousefine := false;
end;

procedure xhidecursor;
begin
  if usemouse then
    hidecursor
  else
    revcursor (mx,my);
end;

procedure xshowcursor;
begin
  if usemouse then
    showcursor
  else
    revcursor (mx,my);
end;

procedure updatefinecalc;
  var
    cxii,cyii : integer;
begin
  if usemouse and mmousefine then
  begin
    cxi := cxi + (mx - mxa);
    cyi := cyi + (my - mya);
    cxii := cxi div mmouseratio;
    cyii := cyi div mmouseratio;
    cx := cxii / 2;
    cy := cyii / 2;
    mx := cxii div 2;
    my := cyii div 2;
    cursor := cursorar[(cxii mod 2) + 2*(cyii mod 2)];
    hidecursor;
    defgcursor (7,7);
    setpos (mx,my);
    showcursor;
  end
  else
  begin
    cx := mx;
    cy := my;
  end;
end;

procedure startfine;
begin
  mmousefine := mousefine;
  if (not mmousefine) or (not usemouse) then
  begin
    mmouseratio := 1;
    exit;
  end;
  if mouseslow then
    mmouseratio := mouseratio
  else
    mmouseratio := 1;
  if ((cxi div mmouseratio) div 2 <> mx) or
     ((cyi div mmouseratio) div 2 <> my) then
  begin
    cxi := mx*2*mmouseratio;
    cyi := my*2*mmouseratio;
    cx := mx;
    cy := my;
  end;
  updatefinecalc;
end;

procedure updatefine (usecurkeys : boolean);
begin
  if usemouse then
    repeat
      getposbut (mx,my,mbl,mbr);
    until (mxa <> mx) or (mya <> my) or
          (mbla <> mbl) or (mbra <> mbr) or keypressed
  else
    repeat
    until keypressed;
  if keypressed and not mkey then
  begin
    mkey := true;
    mc1 := readkey;
    if usecurkeys and (mc1 in [keyup,keydown,keyleft,keyright]) then
    begin
      case mc1 of
        keyup    : my := my - mmouseratio*curkey;
        keydown  : my := my + mmouseratio*curkey;
        keyleft  : mx := mx - mmouseratio*curkey;
        keyright : mx := mx + mmouseratio*curkey;
      end;
      if usemouse then
        setpos (mx,my)
      else
      begin
        limitborder (mx,my);
        revcursor (mxa,mya);
        revcursor (mx,my);
      end;
      mkey := false;
    end;
    if mc1 = #0 then
    begin
      mc2 := readkey;
      if usecurkeys and (mc2 in [curup,curdown,curleft,curright]) then
      begin
        case mc2 of
          curup    : my := my - mmouseratio;
          curdown  : my := my + mmouseratio;
          curleft  : mx := mx - mmouseratio;
          curright : mx := mx + mmouseratio;
        end;
        if usemouse then
          setpos (mx,my)
        else
        begin
          limitborder (mx,my);
          revcursor (mxa,mya);
          revcursor (mx,my);
        end;
        mkey := false;
      end;
    end;
  end;
  updatefinecalc;
  mxa := mx;
  mya := my;
  if not usemouse then
    exit;
  mblold := mbla;
  mbrold := mbra;
  mbla := mbl;
  mbra := mbr;
end;

procedure endfine;
begin
  if (not mmousefine) or (not usemouse) then
    exit;
  mmousefine := false;
  cursor := cursorar[0];
  hidecursor;
  defgcursor (7,7);
  setpos (mx,my);
  showcursor;
end;

function clickedl : boolean;
begin
  clickedl := usemouse and mbl and not mblold;
end;

function clickedr : boolean;
begin
  clickedr := usemouse and mbr and not mbrold;
end;

function anyclick : boolean;
begin
  anyclick := clickedl or clickedr or mkey;
end;

function getmkey : char;
begin
  getmkey := mc1;
  mkey := false;
end;

function clickedlkey : boolean;
begin
  if mkey then
    clickedlkey := (getmkey = ' ')
  else
    clickedlkey := clickedl;
end;

function clickedrkey : boolean;
begin
  if mkey then
    clickedrkey := (getmkey <> ' ')
  else
    clickedrkey := clickedr;
end;

procedure waitclickkey;
begin
  repeat
    updatefine (true);
  until anyclick;
  mkey := false;
end;


{ procedures for basic graphic & text  i/o }

procedure newscreen;
begin
  xhidecursor;
  setfillstyle (1,bgc);
  bar (0,0,xmax,ymax);
  xshowcursor;
end;

procedure box (x1,y1,x2,y2 : integer);
begin
  moveto (x1,y1);
  lineto (x1,y2);
  lineto (x2,y2);
  lineto (x2,y1);
  lineto (x1,y1);
end;

procedure revbox (x1,y1,x2,y2 : integer);
  var
    t : integer;
begin
  if (x1 = x2) or (y1 = y2) then
    revline (x1,y1,x2,y2)
  else
  begin
    if x2 < x1 then
    begin
      t := x1;
      x1 := x2;
      x2 := t;
    end;
    if y2 < y1 then
    begin
      t := y1;
      y1 := y2;
      y2 := t;
    end;
    revline (x1,y1,x1,y2-1);
    revline (x1,y2,x2-1,y2);
    revline (x2,y2,x2,y1+1);
    revline (x2,y1,x1+1,y1);
  end;
end;

procedure errorbeep;
begin
  sound (1000);
  delay (100);
  nosound;
end;

procedure newtext (x,y : integer; xcent : boolean; s : string);
begin
  xhidecursor;
  if xcent then
    x := x - (textwidth(s)+tx1+tx2) div 2;
  bar (x,y,x+textwidth(s)+tx1+tx2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (s);
  xshowcursor;
end;

procedure boxtext (x1,y1,x2,y2 : integer; xcent : boolean; s : string);
begin
  xhidecursor;
  bar (x1,y1,x2,y2);
  box (x1+1,y1+1,x2-1,y2-1);
  if xcent then
    newtext ((x1+x2+1) div 2,y1+3,true,s)
  else
    newtext (x1+3,y1+3,false,s);
  xshowcursor;
end;

function editchar (x,y : integer; que : string; var ans : char) : boolean;
{ <ret>,<key> -> true, <esc> -> false }
  var
    c : char;
    x2 : integer;
begin
  x2 := x+textwidth(que)+tx1;
  xhidecursor;
  bar (x,y,x2+tw+tx2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (que);
  moveto (x2,y+ty1);
  outtext (ans);
  moveto (x2,y+ty1+lowcur);
  outtext ('_');
  xshowcursor;
  repeat
    updatefine (false);
  until clickedr or mkey;
  editchar := false;
  if mkey then
  begin
    c := getmkey;
    if c = cr then
      editchar := true
    else
      if c >= ' ' then
      begin
        ans := c;
        editchar := true;
      end;
  end;
  xhidecursor;
  bar (x2,y,x2+tw+tx2,y+th+ty1+ty2);
  moveto (x2,y+ty1);
  outtext (ans);
  xshowcursor;
end;

function editline (x,y : integer; que : string; var ans : string) : boolean;
{ edit old ans, <ret> to enter new, <esc> to keep old}
{ <ret> -> true, <esc> -> false }
  var
    act : integer;
    c : char;
    old : string;
    x2 : integer;
    w,answ,oldw : integer;
    firstkey : boolean;
begin
  x2 := x+textwidth(que)+tx1;
  xhidecursor;
  bar (x,y,x2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (que);
  xshowcursor;
  old := ans;
  act := 1;
  oldw := 0;
  firstkey := true;
  repeat
    answ := textwidth(ans)+tw;
    w := answ;
    if oldw > w then
      w := oldw;
    oldw := answ;
    xhidecursor;
    bar (x2,y,x2+w+tx2,y+th+ty1+ty2);
    moveto (x2,y+ty1);
    outtext (ans);
    moveto (x2+(act-1)*tw,y+ty1+lowcur);
    outtext ('_');
    xshowcursor;
    repeat
      updatefine (false);
    until anyclick;
    if clickedl then
    begin
      if (mx >= x2) and (mx <= x2+textwidth(ans)+tw) and
         (my >= y+ty1) and (my <= y+th+ty1) then
      begin
        act := trunc((mx-x2)/tw) + 1;
        if act > length(ans)+1 then
          act := length(ans)+1;
      end;
      c := ' ';
    end
    else if clickedr then
      c := esc
    else   { mkey }
    begin
      c := getmkey;
      if c >= ' ' then
      begin
        if firstkey then
        begin
          ans := '';
          act := 1;
        end;
        ans := concat(copy(ans,1,act-1),c,copy(ans,act,length(ans)-act+1));
        inc (act);
      end;
      if c = backsp then
        if act > 1 then
        begin
          ans := concat(copy(ans,1,act-2),copy(ans,act,length(ans)-act+1));
          dec (act);
        end;
      if c = ctrly then
      begin
        ans := '';
        act := 1;
      end;
      if c = #0 then
      begin
        c := mc2;
        case c of
          cleft :  begin
                     if act > 1 then
                       dec (act);
                   end;
          cright : begin
                     if act < length(ans)+1 then
                       inc (act);
                   end;
          chome : act := 1;
          cend : act := length(ans)+1;
          cdel :   begin
                     ans := concat(copy(ans,1,act-1),copy(ans,act+1,length(ans)-act));
                   end;
        end;
        c := ' ';
      end;
    end;
    firstkey := false;
  until c in [cr,esc];
  if c = esc then
    ans := old;
  editline := (c = cr);
end;


{ menu & window handling procedures }

function lowcase (c : char) : char;
begin
  if (c >= 'A') and (c <= 'Z') then
    lowcase := char(ord(c)+32)
  else
    lowcase := c;
end;

procedure initm (x,y : integer; vert : boolean);
begin
  menum.x1 := x;
  menum.y1 := y;
  menum.x2 := menum.x1;
  menum.vertical := vert;
  menum.enum := 0;
  menumaxwidth := 0;
end;

procedure nextm (c : char; s : string20);
begin
  inc (menum.enum);
  if menum.vertical then
  begin
    if tw*length(s)+tx1+tx2 > menumaxwidth then
      menumaxwidth := tw*length(s)+tx1+tx2;
  end
  else
  begin
    menum.ewidth[menum.enum] := tw*length(s)+tx1+tx2+2*txb;
    menum.x2 := menum.x2 + menum.ewidth[menum.enum];
  end;
  menum.key[menum.enum] := c;
  menum.name[menum.enum] := s;
end;

procedure quitm (var mm : menut);
  var
    i,msize : integer;
begin
  menum.eheight := th+ty1+ty2;
  if menum.vertical then
  begin
    for i := 1 to menum.enum do
      menum.ewidth[i] := menumaxwidth;
    menum.x2 := menum.x1 + menumaxwidth + 5;
    menum.y2 := menum.y1 + menum.eheight*menum.enum + 5;
  end
  else
  begin
    menum.x2 := menum.x2 + 5;
    menum.y2 := menum.y1 + menum.eheight + 5;
  end;
  msize := imagesize(menum.x1,menum.y1,menum.x2,menum.y2);
  if msize > menumaxsize then
    menumaxsize := msize;
  menum.background := nil;
  mm := menum;
end;

procedure drawmenu (var m : menut);
  var
    i,x,y : integer;
begin
  xhidecursor;
  if m.background <> nil then
    getimage (m.x1,m.y1,m.x2,m.y2,m.background^);
  bar (m.x1,m.y1,m.x2,m.y2);
  box (m.x1+1,m.y1+1,m.x2-1,m.y2-1);
  x := m.x1 + 3;
  y := m.y1 + 3;
  for i := 1 to m.enum do
  begin
    if m.vertical then
    begin
      moveto (x+tx1,y+ty1);
      y := y + m.eheight;
    end
    else
    begin
      moveto (x+txb+tx1,y+ty1);
      x := x + m.ewidth[i];
    end;
    outtext (m.name[i]);
  end;
  xshowcursor;
end;

function checkmenu (var m : menut) : char;
  var
    c : char;
    i,x,y : integer;
begin
  checkmenu := esc;
  if mkey then
  begin
    c := getmkey;
    if c = #0 then
      case mc2 of
        cpgup : checkmenu := 'O';
        cpgdn : checkmenu := 'I';
        chome : checkmenu := 'M';
        cend  : checkmenu := 'L';
        cins  : checkmenu := 'G';
      end;
    if c = tab then
      checkmenu := 'C'
    else if c = ' ' then
      checkmenu := c
    else
      for i := 1 to m.enum do
        if lowcase(c) = m.key[i] then
          checkmenu := m.key[i];
    exit;
  end;
  if clickedr then
    exit;
  checkmenu := ' ';   { clickedl }
  x := m.x1 + 3;
  y := m.y1 + 3;
  for i := 1 to m.enum do
  begin
    if (mx >= x) and (mx < x+m.ewidth[i]) and
       (my >= y) and (my < y+m.eheight) then
    begin
      checkmenu := m.key[i];
      exit;
    end;
    if m.vertical then
      y := y + m.eheight
    else
      x := x + m.ewidth[i];
  end;
end;

procedure freemenu (var m : menut);
begin
  xhidecursor;
  if m.background <> nil then
    putimage (m.x1,m.y1,m.background^,normalput)
  else
  begin
    bar (m.x1,m.y1,m.x2,m.y2);
    redraw := true;
  end;
  xshowcursor;
end;

procedure initconfirm (minleft : longint);
  var
    x1,y1,x2,y2 : integer;
    xs,ys : integer;
    memneed : longint;
begin
  xs := wincol*tw;
  xs := xs+tx1+tx2 + 2*twb;
  ys := 3*(th+ty1+ty2) + 2*twb;
  x1 := (xmax+1) div 2 - (xs+1) div 2;
  y1 := (ymax+1) div 2 - (ys+1) div 2;
  x2 := x1 + xs - 1;
  y2 := y1 + ys - 1;
  memneed := imagesize(x1,y1,x2,y2);
  if memneed+minleft < maxavail then
    getmem (confirmbackground,memneed)
  else
    confirmbackground := nil;
end;

function confirm (s : string; yesno : boolean) : boolean;
  var
    x1,y1,x2,y2 : integer;
    xs,ys : integer;
    xy,xn,yy : integer;
begin
  s := copy(s,1,wincol);
  xs := length(s)*tw;
  if xs < 9*tw then
    xs := 9*tw;
  xs := xs+tx1+tx2 + 2*twb;
  ys := 3*(th+ty1+ty2) + 2*twb;
  x1 := (xmax+1) div 2 - (xs+1) div 2;
  y1 := (ymax+1) div 2 - (ys+1) div 2;
  x2 := x1 + xs - 1;
  y2 := y1 + ys - 1;
  if yesno then
  begin
    xy := xmax div 2 - 3*tw - (3*tw) div 2;
    xn := xmax div 2 + 3*tw - (2*tw) div 2;
  end
  else
    xy := xmax div 2 - (2*tw) div 2;
  yy := y1 + twb + 2*(th+ty1+ty2) + ty1;
  xhidecursor;
  if confirmbackground <> nil then
    getimage (x1,y1,x2,y2,confirmbackground^);
  bar (x1,y1,x2,y2);
  box (x1+1,y1+1,x2-1,y2-1);
  moveto (xmax div 2 - textwidth(s) div 2,y1 + twb + ty1);
  outtext (s);
  if yesno then
  begin
    moveto (xy,yy);
    outtext ('Yes');
    moveto (xn,yy);
    outtext ('No');
  end
  else
  begin
    moveto (xy,yy);
    outtext ('OK');
  end;
  xshowcursor;
  repeat
    updatefine (true);
  until anyclick;
  if yesno then
  begin
    if mkey then
      confirm := (lowcase(getmkey) = 'y')
    else
      if clickedl then
        confirm := ((mx >= xy-tx1) and (mx < xy+3*tw+tx2) and
                    (my >= yy-ty1) and (my < yy+th+ty2))
      else
        confirm := false;   { clickedr }
  end
  else
  begin
    mkey := false;
    confirm := true;
  end;
  xhidecursor;
  if confirmbackground <> nil then
    putimage (x1,y1,confirmbackground^,normalput)
  else
  begin
    bar (x1,y1,x2,y2);
    redraw := true;
  end;
  xshowcursor;
end;

procedure initwindow (minleft : longint);
  var
    w,h,i : integer;
    memneed : longint;
begin
  w := tw*wincol+tx1+tx2+2*twb;
  h := (th+ty1+ty2)*winrow+2*twb;
  winx1 := xmax div 2 - w div 2;
  winy1 := ymax div 2 - h div 2;
  winx2 := winx1 + w - 1;
  winy2 := winy1 + h - 1;
  memneed := imagesize(winx1,winy1,winx2,winy2);
  if memneed+minleft < maxavail then
    getmem (winbackground,memneed)
  else
    winbackground := nil;
  winxc := (winx1 + winx2) div 2;
  for i := 1 to winrow do
    winyr[i] := winy1 + twb + (i-1)*(th+ty1+ty2);
end;

procedure openwindow;
begin
  xhidecursor;
  if winbackground <> nil then
    getimage (winx1,winy1,winx2,winy2,winbackground^);
  bar (winx1,winy1,winx2,winy2);
  box (winx1+1,winy1+1,winx2-1,winy2-1);
  xshowcursor;
end;

procedure closewindow;
begin
  xhidecursor;
  if winbackground <> nil then
    putimage (winx1,winy1,winbackground^,normalput)
  else
  begin
    bar (winx1,winy1,winx2,winy2);
    redraw := true;
  end;
  xshowcursor;
end;


begin
end.

{ end unit }
