program gpslog;

{
************************************************************

(c) Heiko Purnhagen 1993-1997
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/


All usual disclaimers apply!

This program may be copied according to the GNU General Public Licence.

Please check README for further details.

************************************************************
}


{ (c) Heiko Purnhagen  06-aug-95 }
{ added GPRMC   HP 960719 }


{ important options: b- }
{ suggested options: n- e- r- d- l- }


uses
  crt,dos;

const
  title = 'gpslog (c) Heiko Purnhagen 19-jul-96';

                { 8250 Registers }
  com1 = $3f8;  { com1 base }
  com2 = $2f8;  { com2 base }
  rbr = $00;    { Receiver Buffer Register }
  thr = $00;    { Transmitter Holding Register }
  ier = $01;    { Interrupt Enable Register }
  iir = $02;    { Interrupt Identification Register }
  lcr = $03;    { Line Control Register }
  mcr = $04;    { Modem Control Register }
  lsr = $05;    { Line Status Register }
  msr = $06;    { Modem Status Register }

  br4800  = $0018; { baud rate divisor }
  br9600  = $000c;

                 { parity data stop }
  modeo81 = $0b; {  odd    8    1 }
  moden81 = $03; {  no     8    1 }

  gpsbuflen = 128;
  gpsinittimeout = 32767;

  daysec = 24.0*60.0*60.0;

  nmeainittimeout = 32767;
  nmeastrlen = 255;
  nmeawptlen = 8;

  tllformat = 2;
  tlllength = 24;


type
  nmeawpt = string[nmeawptlen];
  llt = record
          lat,lon : real;
        end;

var
  combase : word;
  oldbaud,oldmode,oldint : word;

  gpsbuffer : array [0..gpsbuflen-1] of byte;
  gpstimeout : integer;
  gpsbufhead : integer;

  nmeatimeout : integer;

  { all gps data in rad, m, m/s }
  gpstime : real;
  gpstimeh,gpstimem,gpstimes : integer;
  gpsdatey,gpsdatem,gpsdated : integer;
  gpsx,gpsy,gpsz : real;
  gpsvele,gpsveln,gpsvelup : real;
  gpslat,gpslon,gpsalt : real;
  gpshdp,gpsvdp,gpspdp,gpstdp,gpsgdp : real;
  gpsehpe,gpsevpe,gpsete,gpsehv : real;
  gpssat : array[1..5] of record
                            track : boolean;
                            prn : integer;
                            dbhz : integer;
                          end;

  ll : llt;
  c : char;
  filename : string;
  f : text;
  gpsnmea : boolean;
  gpscom : integer;
  getwpts : boolean;
  fileok : boolean;
  dt : integer;
  nextlogtime : real;
  gpsbytesleft : integer;


procedure initser (com,baud : word; mode : string);
  var
    dummy,bauddiv,modeb : word;
begin
  if com = 2 then
    combase := com2
  else
    combase := com1;
  if baud = 4800 then
    bauddiv := br4800
  else
    bauddiv := br9600;
  if mode = 'n81' then
    modeb := moden81
  else
    modeb := modeo81;
  inline ($fa);                      { cli }
  oldmode := port[combase+lcr];
  port[combase+lcr] := 00;           { reset DLAB }
  dummy := port[combase+rbr];        { empty receiver register }
  dummy := port[combase+rbr];        { empty receiver register }
  port[combase+lcr] := $80;          { set Divisor Latch Adress Bit (DLAB) }
  oldbaud := port[combase+thr]+
             256*port[combase+ier];
  port[combase+thr] := lo(bauddiv);  { set Divisor Low Byte }
  port[combase+ier] := hi(bauddiv);  { set Divisor High Byte }
  port[combase+lcr] := modeb;        { mode, reset DLAB }
  oldint := port[combase+ier];
  port[combase+ier]:=  $00;          { no interrupts }
  inline($fb)                        { sti }
end;

procedure restoreser;
  var
    dummy : word;
begin
  inline ($fa);                      { cli }
  port[combase+lcr] := 00;           { reset DLAB }
  dummy := port[combase+rbr];        { empty receiver register }
  dummy := port[combase+rbr];        { empty receiver register }
  port[combase+lcr] := $80;          { set Divisor Latch Adress Bit (DLAB) }
  port[combase+thr] := lo(oldbaud);  { set Divisor Low Byte }
  port[combase+ier] := hi(oldbaud);  { set Divisor High Byte }
  port[combase+lcr] := oldmode;      { mode, reset DLAB }
  port[combase+ier] := oldint;       { restore interrupts }
  inline($fb)                        { sti }
end;

function readserok : boolean;
begin
  inline($fa);                                        { cli }
  readserok := ((port[combase+lsr] and $01 ) = $01);  { data received? }
  inline($fb);                                        { sti }
end;

function readser : byte;
begin
  inline($fa);                              { cli }
  repeat
  until (port[combase+lsr] and $01) = $01;  { wait until data received }
  readser := port[combase+rbr];             { read receiver register }
  inline($fb);                              { sti }
end;

function writeserok : boolean;
begin
  writeserok := ((port[combase+lsr] and $20) = $20);  { data transmitted? }
end;

procedure writeser (data : byte);
begin
  inline($fa);                              { cli }
  repeat
  until (port[combase+lsr] and $20) = $20;  { wait until data trasmitted }
  port[combase+thr] := data;                { write transmitter register }
  inline($fb);                              { sti }
end;

procedure initgps (com : integer);
begin
  initser (com,9600,'o81');
  gpstimeout := gpsinittimeout;
  gpsbufhead := 4;
  gpsbuffer[0] := $ff;
  gpsbuffer[1] := $81;
  gpsbuffer[2] := $67;
  gpsbuffer[3] := $00;
end;

function readgps : integer;
{ 0 = ok  1..gpsbuflen = number of bytes left (timeout) }
  var
    timeout,bufidx : integer;
begin
  timeout := 0;
  bufidx := 0;
  repeat
    repeat
      inc (timeout);
    until readserok or (timeout = gpstimeout);
    if readserok then
    begin
      if bufidx < gpsbufhead then
      begin
        if readser = gpsbuffer[bufidx] then
          inc (bufidx)
        else
          bufidx := 0;
      end
      else
      begin
        gpsbuffer[bufidx] := readser;
        inc (bufidx);
      end;
    end;
  until (bufidx = gpsbuflen) or (timeout = gpstimeout);
  if bufidx = gpsbuflen then
    gpstimeout := timeout * 2
  else
    gpstimeout := gpsinittimeout;
  readgps := gpsbuflen-bufidx;
end;

{  (msb) s  m e (lsb) }
{ fp     1 23 8 (4 byte) }
{ efp    1 39 8 (6 byte) }
{ value = -1^s * 2^(e-129) * 1.m }

function getgpsefp (idx : integer) : real;
  var
    r : real;
begin
  r := gpsbuffer[idx+1]/256;
  r := (gpsbuffer[idx+2]+r)/256;
  r := (gpsbuffer[idx+3]+r)/256;
  r := (gpsbuffer[idx+4]+r)/256;
  r := 1+((gpsbuffer[idx+5] and $7f)+r)/128;
  if gpsbuffer[idx+5] and $80 = $80 then
    r := -r;
  getgpsefp := r*exp((gpsbuffer[idx]-129)*ln(2));
end;

function getgpsfp (idx : integer) : real;
  var
    r : real;
begin
  r := gpsbuffer[idx+1]/256;
  r := (gpsbuffer[idx+2]+r)/256;
  r := 1+((gpsbuffer[idx+3] and $7f)+r)/128;
  if gpsbuffer[idx+3] and $80 = $80 then
    r := -r;
  getgpsfp := r*exp((gpsbuffer[idx]-129)*ln(2));
end;

function getgpsword (idx : integer) : word;
{ ls byte first }
begin
  getgpsword := gpsbuffer[idx]+256*gpsbuffer[idx+1];
end;

procedure getgpsdata;
  var
    time : longint;
    i : integer;
begin
  gpstime := getgpsefp(20);
  time := round(gpstime);
  gpstimes := time mod 60;
  time := time div 60;
  gpstimem := time mod 60;
  gpstimeh := time div 60;
  gpsdated := getgpsword(26);
  gpsdatem := getgpsword(28);
  gpsdatey := getgpsword(30);
  gpsx := getgpsefp(36);
  gpsy := getgpsefp(42);
  gpsz := getgpsefp(48);
  gpsvele := getgpsfp(54);
  gpsveln := getgpsfp(58);
  gpsvelup := getgpsfp(62);
  gpslat := getgpsefp(66);
  gpslon := getgpsefp(72);
  gpsalt := getgpsfp(78);
  gpshdp := getgpsword(82) / 100;
  gpsvdp := getgpsword(84) / 100;
  gpspdp := getgpsword(86) / 100;
  gpstdp := getgpsword(88) / 100;
  gpsgdp := getgpsword(90) / 100;
  gpsehpe := getgpsword(112);
  gpsevpe := getgpsword(114);
  gpsete := getgpsword(116) / 300e6;
  gpsehv := getgpsword(118) / 100;
  for i := 1 to 5 do
  begin
    gpssat[i].track := ((gpsbuffer[93+(i-1)*4] and $8) = $8);
    gpssat[i].prn := gpsbuffer[92+(i-1)*4];
    gpssat[i].dbhz := gpsbuffer[95+(i-1)*4];
  end;
end;

function date2str (d,m,y : integer) : string;
  var
    s,ss : string;
begin
  str (d:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := ss+'-';
  ss := '';
  case m of
     1 : ss := 'jan';
     2 : ss := 'feb';
     3 : ss := 'mar';
     4 : ss := 'apr';
     5 : ss := 'may';
     6 : ss := 'jun';
     7 : ss := 'jul';
     8 : ss := 'aug';
     9 : ss := 'sep';
    10 : ss := 'oct';
    11 : ss := 'nov';
    12 : ss := 'dec';
  end;
  s := s+ss+'-';
  str (y:4,ss);
  date2str := s+ss;
end;

function getgpstimestr : string;
  var
    s,ss : string;
begin
  s := date2str(gpsdated,gpsdatem,gpsdatey);
  s := s+' ';
  str (gpstimeh:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+':';
  str (gpstimem:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+':';
  str (gpstimes:2,ss);
  if ss[1]=' ' then
    ss[1] := '0';
  s := s+ss+' ';
  getgpstimestr := s;
end;

{ nmea0183   03-aug-95 }

procedure initnmea (com : integer);
begin
  initser (com,4800,'n81');
  nmeatimeout := nmeainittimeout;
end;

function readnmea (var sent : string) : integer;
{ 0=timeout -1=error -2=check error 1=sentence w/o check 2=sent. w check }
{ check error not implemented yet }
  var
    timeout : integer;
    i,tmp,state : integer;
begin
  readnmea := 0;
  timeout := 0;
  i := 0;
  sent := '';
  state := 0;
  repeat
    repeat
      inc (timeout);
    until readserok or (timeout = nmeatimeout);
    if timeout = nmeatimeout then
      exit;
    tmp := readser;
    if tmp = ord('$') then
    begin
      inc (i);
      sent[i] := char(tmp);
      state := 1;
    end;
  until state = 1;
  repeat
    repeat
      inc (timeout);
    until readserok or (timeout = nmeatimeout);
    if timeout = nmeatimeout then
      exit;
    tmp := readser;
    if i >= nmeastrlen then
    begin
      readnmea := -1;
      exit;
    end
    else
    begin
      if tmp = 13 then   { cr }
        state := 2
      else
      begin
        inc (i);
        sent[i] := char(tmp);
      end;
    end;
  until state = 2;
  { lf is not checked }
  sent[0] := chr(i);
  if i < 6 then
  begin
    readnmea := -1;
    exit;
  end;
  if sent[i-2] = '*' then
  begin
    readnmea := 2;   { sentence with checksum }
    sent[0] := chr(i-3);
  end
  else
    readnmea := 1;   { sentence without checksum }
end;

function getnmeatype (sent : string) : integer;
{ 0=unkown 1=GLL 2=VTG 3=WPL 4=R00 5=RMC }
  var
    t : string[3];
begin
  t := copy(sent,4,3);
  if t='GLL' then
    getnmeatype := 1
  else if t='VTG' then
    getnmeatype := 2
  else if t='WPL' then
    getnmeatype := 3
  else if t='R00' then
    getnmeatype := 4
  else if t='RMC' then
    getnmeatype := 5
  else
    getnmeatype := 0;
end;

function getnmeall (sent : string; var idx : integer;
                    var lat,lon : real) : boolean;
{ OK->true error->false }
  var
    hh,mm : real;
    c,i : integer;
begin
  getnmeall := false;
  if length(sent) < idx+1 then
    exit;
  val(copy(sent,idx,2),hh,c);
  if c <> 0 then
    exit;
  idx := idx+2;
  i := idx;
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  val(copy(sent,idx,i-idx),mm,c);
  if c <> 0 then
    exit;
  idx := i+1;
  if length(sent) < idx then
    exit;
  if (sent[idx] <> 'N') and (sent[idx] <> 'S') then
    exit;
  lat := (hh+mm/60)/180*pi;
  if sent[idx] = 'S' then
    lat := -lat;
  inc (idx);
  if (length(sent) < idx) or (sent[idx] <> ',') then
    exit;
  inc (idx);
  if length(sent) < idx+2 then
    exit;
  val(copy(sent,idx,3),hh,c);
  if c <> 0 then
    exit;
  idx := idx+3;
  i := idx;
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  val(copy(sent,idx,i-idx),mm,c);
  if c <> 0 then
    exit;
  idx := i+1;
  if length(sent) < idx then
    exit;
  if (sent[idx] <> 'E') and (sent[idx] <> 'W') then
    exit;
  lon := (hh+mm/60)/180*pi;
  if sent[idx] = 'W' then
    lon := -lon;
  inc (idx);
  getnmeall := true;
end;

function getnmeagll (sent : string; var lat,lon : real) : boolean;
{ OK->true error->false }
  var
    i : integer;
begin
  getnmeagll := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  getnmeagll := getnmeall(sent,i,lat,lon);
end;

function getnmeavtg (sent : string; var crs,spd : real) : boolean;
{ OK->true error->false }
  var
    i,j,c : integer;
begin
  getnmeavtg := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  if j = i then
    crs := 0
  else
  begin
    val(copy(sent,i,j-i),crs,c);
    if c <> 0 then
      exit;
    crs := crs/180*pi;
  end;
  i := j+1;
  if length(sent) < i then
    exit;
  if sent[i] <> 'T' then
    exit;
  inc (i);
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  { skip mag crs }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  { skip M }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  val(copy(sent,i,j-i),spd,c);
  if c <> 0 then
    exit;
  i := j+1;
  if length(sent) < i then
    exit;
  if sent[i] <> 'N' then
    exit;
  getnmeavtg := true;
end;

function getnmeawpl (sent : string; var lat,lon : real;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
  var
    i,j : integer;
begin
  getnmeawpl := false;
  name := '';
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  if not getnmeall(sent,i,lat,lon) then
    exit;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  j := j-i;
  if j > nmeawptlen then
    j := nmeawptlen;
  name := copy(sent,i,j);
  getnmeawpl := true;
end;

function getnmear00 (sent : string; var idx : integer;
                     var name : nmeawpt) : boolean;
{ OK->true error->false }
{ in: idx=wpt index (1,2,...)   out: idx=num. of wtps }
  var
    i,j,k,n : integer;
begin
  getnmear00 := false;
  name := '';
  i := 7;
  n := 0;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  repeat
    inc (n);
    j := i;
    while (j <= length(sent)) and (sent[j] <> ',') do
      inc (j);
    if n = idx then
    begin
      k := j-i;
      if k > nmeawptlen then
        k := nmeawptlen;
      name := copy(sent,i,k);
      getnmear00 := true;
    end;
    i := j;
    if (i > length(sent)) then
    begin
      idx := n;
      exit;
    end;
    inc (i);
    if (i > length(sent)) or (sent[i] = ',') then
    begin
      idx := n;
      exit;
    end;
  until false;
end;

function getnmearmc (sent : string; var lat,lon,crs,spd : real) : boolean;
{ OK->true error->false }
  var
    i,j,c : integer;
begin
  getnmearmc := false;
  i := 7;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  { skip hhmmss }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  { skip A }
  while (i <= length(sent)) and (sent[i] <> ',') do
    inc (i);
  if i > length(sent) then
    exit;
  inc (i);
  if not getnmeall(sent,i,lat,lon) then
    exit;
  if (length(sent) < i) or (sent[i] <> ',') then
    exit;
  inc (i);
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  val(copy(sent,i,j-i),spd,c);
  if c <> 0 then
    exit;
  i := j+1;
  if length(sent) < i then
    exit;
  j := i;
  while (j <= length(sent)) and (sent[j] <> ',') do
    inc (j);
  if j > length(sent) then
    exit;
  if j = i then
    crs := 0
  else
  begin
    val(copy(sent,i,j-i),crs,c);
    if c <> 0 then
      exit;
    crs := crs/180*pi;
  end;
  { ignore ddmmyy & deviation }
  getnmearmc := true;
end;

{ lat,lon <-> string }

function ang2str (ang : real; h3 : boolean; len : integer) : string;
  var
    h,m,s : real;
    sh,sm,ss : string;
    l,h1,h2,m1,m2,s1,s2 : integer;
begin
  h := ang/pi*180;
  sm := '';
  ss := '';
  if h3 then
    l := 3
  else
    l := 2;
  h1 := l;
  h2 := 0;
  if tllformat = 1 then
  begin
    if len >= l+3 then
    begin
      h1 := len-1;
      h2 := len-l-2;
    end;
  end
  else
  begin
    m := frac(h)*60.0;
    h := trunc(h);
  end;
  if tllformat <> 1 then
  begin
    len := len-l-1;
    m1 := 2;
    m2 := 0;
    if tllformat <> 3 then
    begin
      if len >= 5 then
      begin
        m1 := len-1;
        m2 := len-4;
      end;
    end
    else
    begin
      s := frac(m)*60.0;
      m := trunc(m);
    end;
    if tllformat = 3 then
    begin
      len := len-3;
      s1 := 2;
      s2 := 0;
      if len >= 5 then
      begin
        s1 := len-1;
        s2 := len-4;
      end;
      str (s:s1:s2,ss);
      if ss[1] = '6' then
      begin
        str (0.0:s1:s2,ss);
        m := m+1;
      end;
      if ss[1] = ' ' then
        ss[1] := '0';
      ss := ss+'"';
    end;
    str (m:m1:m2,sm);
    if sm[1] = '6' then
    begin
      str (0.0:m1:m2,sm);
      h := h+1;
    end;
    if sm[1] = ' ' then
      sm[1] := '0';
    sm := sm+'''';
  end;
  str (h:h1:h2,sh);
  if l = 3 then
    if sh[2] = ' ' then
      sh[2] := '0';
  if sh[1] = ' ' then
    sh[1] := '0';
  sh := sh+'';
  ang2str := sh+sm+ss;
end;

function ll2str (ll : llt) : string;
  var
    s,ss : string;
    lt : real;
    len : integer;
begin
  len := tlllength div 2;
  if ll.lat >= 0 then
  begin
    lt := ll.lat;
    ss := 'N';
  end
  else
  begin
    lt := -ll.lat;
    ss := 'S';
  end;
  s := ang2str(lt,false,len-2)+ss;
  if ll.lon >= 0 then
  begin
    lt := ll.lon;
    ss := 'E';
  end
  else
  begin
    lt := -ll.lon;
    ss := 'W';
  end;
  ll2str := s+' '+ang2str(lt,true,len-1)+ss;
end;

function str2ang (s : string; var ang : real) : boolean;
{ OK -> true, ERROR -> false}
  const
    tab = #9;
  var
    ts : string;
    hh,mm,ss : real;
    p,c : integer;
begin
  str2ang := false;
  hh := 0;
  mm := 0;
  ss := 0;
  while (s <> '') and (s[length(s)] in [' ',tab]) do
    delete (s,length(s),1);
  while (s <> '') and (s[1] in [' ',tab]) do
    delete (s,1,1);
  if s = '' then
    exit;
  p := pos('',s);
  if p = 0 then
    p := pos(' ',s);
  if p = 0 then
    ts := s
  else
    ts := copy(s,1,p-1);
  val (ts,hh,c);
  if c <> 0 then
    exit;
  if hh < 0 then
    exit;
  if p > 0 then
  begin
    delete (s,1,p);
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    if s <> '' then
    begin
      p := pos('''',s);
      if p = 0 then
        p := pos(' ',s);
      if p = 0 then
        ts := s
      else
        ts := copy(s,1,p-1);
      val (ts,mm,c);
      if c <> 0 then
        exit;
      if (frac(hh) <> 0) or (mm < 0) or (mm > 60) then
        exit;
      if p > 0 then
      begin
        delete (s,1,p);
        while (s <> '') and (s[1] in [' ',tab]) do
          delete (s,1,1);
        if s <> '' then
        begin
          p := pos('"',s);
          if p = 0 then
            p := pos(' ',s);
          if p = 0 then
            ts := s
          else
            ts := copy(s,1,p-1);
          val (ts,ss,c);
          if c <> 0 then
            exit;
          if (frac(mm) <> 0) or (ss < 0) or (ss > 60) then
            exit;
          if p > 0 then
          begin
            delete (s,1,p);
            while (s <> '') and (s[1] in [' ',tab]) do
              delete (s,1,1);
            if s <> '' then
              exit;
          end;
        end;
      end;
    end;
  end;
  ang := (hh+mm/60+ss/3600)/180*pi;
  str2ang := true;
end;

function str2ll (s : string; var ll : llt) : boolean;
{ OK -> true, ERROR -> false}
  var
    ss : string;
    sign : integer;
    p1,p2 : integer;
begin
  str2ll := false;
  p1 := pos('N',s);
  if p1 = 0 then
    p1 := pos('n',s);
  sign := 1;
  if p1 = 0 then
  begin
    p1 := pos('S',s);
    if p1 = 0 then
      p1 := pos('s',s);
    sign := -1;
  end;
  if p1 = 0 then
    exit;
  if not str2ang(copy(s,1,p1-1),ll.lat) then
      exit;
  ll.lat := sign*ll.lat;
  if (ll.lat > pi/2) or (ll.lat < -pi/2) then
    exit;
  p2 := pos('E',s);
  if p2 = 0 then
    p2 := pos('e',s);
  sign := 1;
  if p2 = 0 then
  begin
    p2 := pos('W',s);
    if p2 = 0 then
      p2 := pos('w',s);
    sign := -1;
  end;
  if p2 = 0 then
    exit;
  if not str2ang(copy(s,p1+1,p2-p1-1),ll.lon) then
      exit;
  ll.lon := sign*ll.lon;
  if (ll.lon > 2*pi) or (ll.lon < -2*pi) then   { avoids problems at 180 }
    exit;
  str2ll := true;
end;


{ main procedure }

procedure readc (var c : char);
  var
    s : string;
begin
  readln (s);
  if length(s) = 1 then
    c := s[1]
  else
    c := ' ';
end;

procedure gettimedate;
  var
    t1,t2,t3,t4 : word;
begin
  getdate (t1,t2,t3,t4);
  gpsdatey := t1;
  gpsdatem := t2;
  gpsdated := t3;
  gettime (t1,t2,t3,t4);
  gpstimeh := t1;
  gpstimem := t2;
  gpstimes := t3;
  gpstime := gpstimeh;
  gpstime := gpstime*60+gpstimem;
  gpstime := gpstime*60+gpstimes;
end;

procedure getnmeawpts;
  var
    gotall : boolean;
    s : string;
    i : integer;
    numwpts : integer;
    name,name1 : nmeawpt;
    lat,lon : real;
begin
  writeln ('press any key to abort');
  {$i-}
  writeln (f,'. gpslog waypoints');
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t write file');
    exit;
  end;
  gotall := false;
  numwpts := 0;
  repeat
    repeat
    until keypressed or readserok;
    if readserok then
    begin
      i := readnmea (s);
      if i > 0 then
      begin
        i := getnmeatype(s);
        if i = 3 then
          if getnmeawpl (s,lat,lon,name) then
          begin
            if numwpts = 0 then
              name1 := name;
            if (numwpts > 0) and (name = name1) then
              gotall := true
            else
            begin
              ll.lat := lat;
              ll.lon := lon;
              s := ll2str(ll);
              {$i-}
              writeln (f,'mark');
              {$i+}
              if ioresult <> 0 then
              begin
                writeln ('error: can''t write file');
                exit;
              end;
              {$i-}
              writeln (f,name);
              {$i+}
              if ioresult <> 0 then
              begin
                writeln ('error: can''t write file');
                exit;
              end;
              {$i-}
              writeln (f,s);
              {$i+}
              if ioresult <> 0 then
              begin
                writeln ('error: can''t write file');
                exit;
              end;
              inc (numwpts);
              writeln (numwpts,': ',name,' ',s);
            end;
          end;
        end;
    end;
  until gotall or keypressed;
  if gotall then
    writeln ('received all ',numwpts,' waypoints')
  else
    writeln ('aborted: received ',numwpts,' waypoints');
end;

procedure logloop;
  var
    bmark : boolean;
    i : integer;
    r1,r2,r3,r4 : real;
    s,ss : string;
    c : char;
begin
  bmark := false;
  gettimedate;
  nextlogtime := gpstime;
  {$i-}
  writeln (f,'. gpslog log');
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t write file');
    exit;
  end;
  {$i-}
  writeln (f,'line');
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t write file');
    exit;
  end;
  ss := getgpstimestr;
  {$i-}
  writeln (f,'gpslog start ',ss,' dt=',dt,' sec');
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t write file');
    exit;
  end;
  writeln ('logging ...');
  writeln (ss);
  writeln ('press any key to interrupt');
  s := '';
  repeat
    repeat
    until keypressed or readserok;
    if readserok then
    begin
      if gpsnmea then
      begin
        i := readnmea (ss);
        if i > 0 then
        begin
          i := getnmeatype(ss);
          if i = 1 then
          begin
            if not getnmeagll (ss,r1,r2) then
              i := -1
            else
            begin
              gpslat := r1;
              gpslon := r2;
            end;
          end
          else if i= 5 then
          begin
            if not getnmearmc (ss,r1,r2,r3,r4) then
              i := -1
            else
            begin
              { writeln(r1,r2,r3/pi*180,r4); }
              gpslat := r1;
              gpslon := r2;
            end;
          end;
        end;
      end
      else   { gpspac }
      begin
        gpsbytesleft := readgps;
        if gpsbytesleft = 0 then
        begin
          i := 1;
          getgpsdata;
        end
        else
          i := -1;
      end;
      if i = 1 then
      begin
        gettimedate;
        ll.lat := gpslat;
        ll.lon := gpslon;
        ss := ll2str (ll);
        if bmark then
          write ('\ ')
        else
          write ('/ ');
        bmark := not bmark;
        writeln (ss);
        if gpstime >= nextlogtime then
        begin
          nextlogtime := nextlogtime + dt;
          { to avoid midnight-problems }
          if nextlogtime > daysec-dt then
            nextlogtime := nextlogtime - daysec;
          writeln ('log');
          {$i-}
          writeln (f,ss);
          {$i+}
          if ioresult <> 0 then
          begin
            writeln ('error: can''t write file');
            exit;
          end;
        end;
      end;
      if i = -1 then
        writeln ('gps error');
    end;
    if keypressed then
    begin
      repeat
        c := readkey;
      until not keypressed;
      writeln ('enter note or q to quit');
      write ('? ');
      readln (s);
      if (s <> 'q') and (s <> 'Q') then
      begin
        gettimedate;
        ss := getgpstimestr;
        writeln (ss);
        {$i-}
        writeln (f,'. ',ss);
        {$i+}
        if ioresult <> 0 then
        begin
          writeln ('error: can''t write file');
          exit;
        end;
        if s <> '' then
        begin
          {$i-}
          writeln (f,'. ',s);
          {$i+}
          if ioresult <> 0 then
          begin
            writeln ('error: can''t write file');
            exit;
          end;
        end;
        nextlogtime := gpstime;
      end;
    end;
  until (s = 'q') or (s = 'Q');
  writeln ('log finished');
end;


{ main }

begin
  writeln (title);
  writeln;
  repeat
    write ('gps device (n=nmea0183 g=gpspac) ? ');
    readc (c);
  until c in ['n','N','g','G'];
  gpsnmea := c in ['n','N'];
  repeat
    write ('interface (1=com1 2=com2) ? ');
    readc (c);
  until c in ['1','2'];
  if c = '1' then
    gpscom := 1
  else
    gpscom := 2;
  if gpsnmea then
  begin
    repeat
      write ('get nmea waypoints (y,n) ? ');
      readc (c);
    until c in ['y','Y','n','N'];
    getwpts := c in ['y','Y'];
  end
  else
    getwpts := false;
  if not getwpts then
  repeat
    write ('log interval [sec] ? ');
    readln (dt);
  until dt > 0;
  repeat
    if getwpts then
      write ('waypoint filename ? ')
    else
      write ('log filename ? ');
    readln (filename);
    if filename = '' then
    begin
      writeln ('log aborted');
      exit;
    end;
    assign (f,filename);
    {$i-}
    reset (f);
    {$i+}
    if ioresult = 0 then
    begin
      repeat
        write ('file exists! overwrite (y,n) ? ');
        readln (c);
      until c in ['y','Y','n','N'];
      fileok := c in ['y','Y'];
    end
    else
      fileok := true;
  until fileok;
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t open file');
    exit;
  end;

  if gpsnmea then
    initnmea (gpscom)
  else
    initgps (gpscom);
  if getwpts then
    getnmeawpts
  else
    logloop;
{  restoreser; doesn't work on hp95lc }
  {$i-}
  close (f);
  {$i+}
  if ioresult <> 0 then
  begin
    writeln ('error: can''t close file');
    exit;
  end;
end.



